#include "Panel_optimize.h"

#include "Log.h"
#include "Text.h"

#include "Main_window.h"
#include "Command.h"
#include "Timer.h"
#include "Lock.h"
//#include "Interaction.h"
#include "Model.h"
#include "Model_kit.h"
#include "Model_adapter.h"
#include "Min.h"

#include <QCloseEvent>

namespace MM
{   

Panel_optimize::
Panel_optimize ()
:
    Panel ("Optimize", "left"),
    player_(Main_window::singleton()),
    minimizer_(0)
{
    setupUi (this);

    doubleSpinBox_Gradient  ->setValue (1e-3);
    doubleSpinBox_Step      ->setValue (1e-2);
    doubleSpinBox_Functional->setValue (1e-3);
    spinBox_Iterations      ->setValue (100);

    connect(checkBox_Advanced,  SIGNAL(toggled (bool)),
            this,               SLOT(set_advanced(bool)));
       
    set_advanced (false);
    //renew ();
}

void Panel_optimize::
set_stop_state ()
{
    player_->player_on_ = false;

    player_->toolButton_Stop   -> setEnabled (false);
    //player_->toolButton_Tune   -> setEnabled (true);
    player_->toolButton_Run    -> setEnabled (true);
    //player_->toolButton_Start  -> setEnabled (true);

    if (this->isHidden ())
        turn (false);
}

void Panel_optimize::
set_run_state ()
{
    player_->player_on_ = true;

    player_->toolButton_Stop   -> setEnabled (true);
    //player_->toolButton_Tune   -> setEnabled (false);
    player_->toolButton_Run    -> setEnabled (false);
    //player_->toolButton_Start  -> setEnabled (false);
}

void Panel_optimize::
turn (bool on)
{
    //MD_simulator & simulator = model().kit().current_MD();

    //if (minimizer_ == 0) //  !simulator.is_running()
    if (!player_->player_on_)
    {
        player_->toolButton_Stop   -> disconnect ();
        player_->toolButton_Back   -> disconnect ();
        player_->toolButton_Tune   -> disconnect ();
        player_->toolButton_Run    -> disconnect ();
        player_->toolButton_Start  -> disconnect ();
        player_->toolButton_Forward-> disconnect ();
        player_->toolButton_Pause  -> disconnect ();

        if (on)
        {
            player_->tabWidget_bottom->setCurrentWidget (player_->tab_Run);

            connect (player_->toolButton_Stop,   SIGNAL(clicked()), this,
                                 SLOT  (stop   ()));
            connect (player_->toolButton_Run,    SIGNAL(clicked()), this,
                                 SLOT  (run    ()));
            set_stop_state ();
        }
        else 
        {
            player_->toolButton_Stop   -> setEnabled (false);
            player_->toolButton_Back   -> setEnabled (false);
            player_->toolButton_Tune   -> setEnabled (false);
            player_->toolButton_Run    -> setEnabled (false);
            player_->toolButton_Start  -> setEnabled (false);
            player_->toolButton_Forward-> setEnabled (false);
            player_->toolButton_Pause  -> setEnabled (false);
        }
    }
    //else
    //    set_run_state();

    //if (on)
    //    renew ();
}

void Panel_optimize::
closeEvent (QCloseEvent * event)
{
    //MD_simulator & simulator = model().kit().current_MD();

    //if (simulator.is_running())
    if (minimizer_ != 0)
        event->ignore ();
}

//void Panel_optimize::
//set_method (const QString & text)
//{
//    //if (text == "Ascalaph MD")
//    //    model().kit().set_MD (model().kit().MD ());
//    //else if (text == "MDynaMix")
//    //    model().kit().set_MD (model().kit().MDynaMix ());
//
//    //renew ();
//}


//void Panel_dynamics::
//renew_current_model (Model &)
//{
//    renew ();
//}
//
//void Panel_dynamics::
//renew (Model_event::Hint /*hint*/)
//{
//    MD_simulator & simulator = model().kit().current_MD();
//
//
//    Text const & tool_name = model().kit().current_MD().title();
//
//    if (tool_name == "MD")
//        Tool->setCurrentIndex (Tool->findText("Ascalaph MD"));
//
//    else if (tool_name == "MDynaMix")
//        Tool->setCurrentIndex (Tool->findText("MDynaMix"));
//    else
//        FLAW (tool_name + " - no such tool.");
//
//
//    bool support_barostat = simulator.support_barostat();
//
//    if (support_barostat)
//    {
//        groupBox_Barostat->setEnabled (true);
//        doubleSpinBox_Pressure->setValue (simulator.desired_pressure().atm());
//
//        bool has_barostat = simulator.has_barostat();
//
//        checkBox_Barostat->setChecked (has_barostat);
//
//        if (has_barostat)
//        {
//            doubleSpinBox_Pressure->setEnabled (true);
//            checkBox_Barostat->setChecked (true);
//        }
//        else
//        {
//            doubleSpinBox_Pressure->setEnabled (false);
//            checkBox_Barostat->setChecked (false);
//        }
//
//        doubleSpinBox_Pressure->setValue (simulator.desired_pressure().atm());
//   }
//    else
//        groupBox_Barostat->setEnabled (false);
//
//
//    bool has_thermostat = simulator.has_thermostat();
//
//    checkBox_Thermostat->setCheckState 
//        (has_thermostat ? Qt::Checked : Qt::Unchecked);
//
//    doubleSpinBox_Temperature->setEnabled    (has_thermostat);
//
//    if (has_thermostat)
//    {
/////        Text type = MD_simulator::current().thermostat().title ();
//        //to_user().status (type);
//        doubleSpinBox_Temperature->setValue (simulator.desired_temperature ().K());
//    }
//
//    //fix to unit viewer
//    doubleSpinBox_Temperature->setValue 
//        (simulator.desired_temperature().K());
//
//    doubleSpinBox_Time_per_long_step->blockSignals (true);
//    doubleSpinBox_Duration->blockSignals (true);
//
//    Duration_unit duration = simulator.duration();
//    doubleSpinBox_Time_per_long_step->setValue (duration.step_length_ps() * 1000.);
//    doubleSpinBox_Duration->          setValue (duration.ps());
//
//    doubleSpinBox_Duration->blockSignals (false);
//    doubleSpinBox_Time_per_long_step->blockSignals (false);
//}

void Panel_optimize::
run ()
{
    struct Process_events : public Command
    {
        int i_;

        Process_events () : i_(0) {}

        void execute () 
        {
            if (++i_ % 1 == 0)
                qApp->processEvents(QEventLoop::ExcludeSocketNotifiers);
        }
    };
    //simulator.after_each_step (new Process_events);

    Timer timer;
    Text  time;
    minimizer_ = 0;

    Model_adapter adapter (model());

    try
    {
        Model_kit::Lock lock_panel (model().kit());

        Gradient_minimizer minimize (adapter, adapter, 1, 1, 1, 
            doubleSpinBox_Bracketing->value());
        minimizer_ = &minimize;

        if (minimizer_ == 0)
            return;

        minimizer_->stop_logic_ = comboBox_Stop_logic->currentText() == "And" 
            ? Minimizer::AND : Minimizer::OR;   

        if (comboBox_Line_search->currentText() == "Cubic")
            minimizer_->line_search_type_ = Minimizer::CUBIC;   
        else if (comboBox_Line_search->currentText() == "Brent")
            minimizer_->line_search_type_ = Minimizer::BRENT; 
        else
            FLAW ("Wrong Line search type.");

        if (model().kit().interaction().common().ad_hoc())
            minimizer_->line_search_type_ = Minimizer::BRENT; 

        minimizer_->set_brack_K        (1.618034);        //GOLD //fix

        minimizer_->set_gradient_tolerance (doubleSpinBox_Gradient  ->value());
        minimizer_->set_scale              (doubleSpinBox_Scale     ->value());

        if (checkBox_Advanced->isChecked())
        {
            minimizer_->set_step_tolerance      
                (doubleSpinBox_Step      ->value());
            minimizer_->set_functional_tolerance
                (doubleSpinBox_Functional->value());
        }
        else
        {
            minimizer_->set_step_tolerance      
                (doubleSpinBox_Gradient->value()/10);
            minimizer_->set_functional_tolerance
                (doubleSpinBox_Gradient->value()/100);
        }

        Minimizer::Type type;

        if      (method->currentText() == "Hybrid LS and CD") 
            type = Minimizer::Hybrid_Lui_Storey_and_Conjugate_Descent;

        else if (method->currentText() == "Hybrid Hu Storey")
            type = Minimizer::Hu_Storey;

        else if (method->currentText() == "Hybrid Dai Yuan zero")
            type = Minimizer::Hybrid_Dai_Yuan_zero;

        else if (method->currentText() == "HS - Hestenes Stiefel")
            type = Minimizer::Hestenes_Stiefel;

        else if (method->currentText() == "LS - Lui Storey")
            type = Minimizer::Lui_Storey;

        else if (method->currentText() == "PR - Polak Ribiere")
            type = Minimizer::Polak_Ribiere;

        else if (method->currentText() == "FR - Fletcher Reeves")
            type = Minimizer::Fletcher_Reeves;

        else if (method->currentText() == "CD - Conjugate Descent")
            type = Minimizer::Conjugate_Descent;

        else if (method->currentText() == "DY - Dai Yuan")
            type = Minimizer::Dai_Yuan;

        else if (method->currentText() == "Steepest Descent")
            type = Minimizer::Steepest_Descent;

        else
        {
            Text message ("Wrong simulation method - ");
            message += method->currentText();
            FLAW (message);
        }

        to_user().status ("Optimization Started");
        set_run_state ();
        //Lock <Interaction> lock (model().kit().interaction ());
        
        try
        {
            //simulator.start ();
            minimizer_->conjugate_gradient (type, spinBox_Iterations->value());
        }
        catch(Text & message)   
            {stop (); to_user().error (message);}
        catch(Flaw & exception) 
            {stop (); to_user().error (exception.message());}
        catch(Contract_violation & exc) 
            {stop (); to_user().error (exc.message());}
        catch(...)              
            {stop (); to_user().fatal_error ("Unknown error.");}

        model().kit().visual().renew ();
        set_stop_state ();

        //model().kit().interaction ().report (&time, "F speed");

        double E = minimizer_->func_val ();

        char stop_reason[100];
        minimizer_->get_stop_reason (stop_reason);

        Text text = Text() 
            + " E = " + E                           + "     " 
    //                + " grad = " + rms_gradient + "     " 
            + " Grad = " + minimizer_->relative_gradient() + "     " 
            + " Iter = " + minimizer_->get_iteration()     + "     "
            + " " + stop_reason                     + "     "
            + timer.elapsed() + " sec   ";
        to_user(). status (text);
     
        //energy.flush_cache ();//fix remove

        text = model().kit().file_name ();
        text += "\n";
        text += "Energy ";      text += Text(E, 15);
        text +=  "  Gradient "; text += Text(minimizer_->relative_gradient   (), 3);
        text += " Step ";       text += Text(minimizer_->step_tolerance      (), 3);
        text += " Function ";   text += Text(minimizer_->functional_tolerance(), 3);
        text += " Iteration ";  text += Text(minimizer_->get_iteration       (), 3);
        text += "\n";
        text += stop_reason;     text += "   "; 
        text += timer.elapsed(); text += " sec\n";
        
        text += "Call: ";
        text += minimizer_->get_function_calls ();         text += " function, "   ;
        text += minimizer_->get_gradient_calls ();         text += " gradient, "   ;
        text += minimizer_->get_bracketing_func_calls ();  text += " bracketing, " ;
        text += minimizer_->get_brent_func_calls ();       text += " Brent, "      ;
        text += minimizer_->get_quadratic_func_calls ();   text += " quadratic, "  ;
        text += minimizer_->get_cubic_func_calls ();       text += " cubic, "      ;
        text += minimizer_->get_other_func_calls ();       text += " others. "     ;

        log() << text << "\n";

    }
    catch(Text & message)           {to_user().error (message);}
    catch(Flaw & exception)         {to_user().error (exception.message());}
    catch(Contract_violation & exc) {to_user().error (exc.message());}
    catch(...)                      {to_user().fatal_error ("Unknown error.");}

    minimizer_ = 0;
}


void Panel_optimize::
stop ()
{
    //MD_simulator & simulator = model().kit().current_MD();
    //simulator.stop ();
    minimizer_->stop();
}

void Panel_optimize::
set_advanced (bool on)
{
    if (on)
    {
        comboBox_Stop_logic     ->show();
        label_Step              ->show();
        doubleSpinBox_Step      ->show();
        label_Functional        ->show();
        doubleSpinBox_Functional->show();
        label_Scale             ->show();
        doubleSpinBox_Scale     ->show();
        label_Bracketing        ->show();
        doubleSpinBox_Bracketing->show();
        label_Line_search       ->show();
        comboBox_Line_search    ->show();
    }
    else
    {
        comboBox_Stop_logic     ->hide();
        label_Step              ->hide();
        doubleSpinBox_Step      ->hide();
        label_Functional        ->hide();
        doubleSpinBox_Functional->hide();
        label_Scale             ->hide();
        doubleSpinBox_Scale     ->hide();
        label_Bracketing        ->hide();
        doubleSpinBox_Bracketing->hide();
        label_Line_search       ->hide();
        comboBox_Line_search    ->hide();
    }
}

}//MM
